# Copyright (c) 2007 Red Hat
# Written by Jakub Hrozek
# -*- coding: UTF8 -*-

import errno
import fnmatch
import selinux
import os
import exceptions
from ConfigParser import NoOptionError

import output 
import mailoutput 
import tuierrors

from scheduler import scheduler
from scheduler import xmloutput
from scheduler.i18n import _

def matches_glob(glob, test):
    """ Yum-like glob matching """
    return fnmatch.fnmatch(test.dsc['header']['name'], glob)

class ST_Action(object):
    """ A skeleton class: One of supported SecTool actions """
    name = "Base class"

    def __init__(self, options, args, formatter=output.StdoutFormatter()):
        self.options = options
        self.args = args
        self.formatter = formatter
        self.loader = scheduler.ST_TestLoader()
        self.suite  = scheduler.ST_TestSuite()

    # generic functions for checking various common prerequisities
    def checkRoot(self):
        """ Some actions require root """
        if os.getuid() != 0:
            raise tuierrors.PreReqError(_("You need to be root to perform %s") % (self.name))


    def checkList(self):
        """ Some actions require a list of arguments to proceed """
        if len(self.args) == 0:
            raise tuierrors.PreReqError(_("Need a list of tests to do %s") % (self.name))

    def checkSELinux(self):
        """ Some actions require SELinux on """
        if selinux.is_selinux_enabled() == 0:
            raise tuierrors.PreReqError(_("You need to have SELinux enabled to perform %s") % (self.name))

    def extractPathFromArgs(self, arg_list):
        """ Walks through parameter list stored in args, filters out
        arguments containing a path separator from the other arguments """
        files = list()
        args = list()

        for test in arg_list:
            if os.path.sep in test:
                files.append(test)
            else:
                args.append(test)  

        return files, args

    def filter_groups(self, test_suite, groups):
        """ Takes a ST_TestSuite, returns the suite filtered by groups """
        if groups != []:
            set_groups = set(groups)
            test_suite.filterInPlace(lambda test, groups: len(test.dsc.groups & groups) != 0, groups=set_groups)

    def load_tests(self, load_path):
        """ Wrapper around ST_TestsLoader::loadFromPath, catches/raises exceptions etc. """
        try:
            self.suite = self.loader.loadFromPath(load_path)
        except exceptions.Exception, e: # FIXME - catch the exceptions namely, one by one
            raise tuierrors.SecToolActionError(self.name, e)

        return self.suite

    def preReq(self):
        " gets called _before_ the actual action; checks for prereqs "
        pass

    def tearDown(self):
        " gets called _after_ the actual action; cleans up "
        pass

    def doStuff(self):
        " this is the actual workhorse..meant to be subclassed "
        pass

    def doAction(self):
        " calls the methods in the right order. Meant to be directly called from the user "
        self.preReq()
        try:
            self.doStuff()
        finally:
            self.tearDown()

class ST_RunAction(ST_Action):
    """ Intermediate class which serves as a parent for all classes that
    actually run some tests """
    name = _("base run class")

    def __init__(self, options, formatter=output.StdoutFormatter()):
        ST_Action.__init__(self, options.options, options.args, formatter)
        self.scheduler_opts = options.scheduler_opts
        self.environ = dict([ (key.upper(), options.scheduler_opts['PATHS'][key]) for key in options.scheduler_opts['PATHS'] ])
        self.xml_output = None
        self.mail = None
        self.report = output.RunReport()

    def _handle_missing_dependency(self, exception):
        result = scheduler.ST_TestResult()
        result.return_code = 0
        result.addError(0, str(exception))

        return result

    def _handle_missing_requirement(self, exception):
        return self._handle_missing_dependency(exception)

    def __run_one_test(self, test, level=scheduler.ST_TestCase.default_level):
        """
        Runs one test on a specified level 

        @param test: the test to run
        @type test: ST_TestCase
        @param level: the level to run at (defaults to scheduler.ST_TestCase.default_level)
        @type level: int
        @raises tuierrors.SecToolActionError: on a fatal error
        @rtype: ST_TestResult
        """
        try:
            result = test.runTestLevel(level)
        except scheduler.errors.MissingDependencyError, e:
            result = self._handle_missing_dependency(e)
        except scheduler.errors.MissingRequirementError, e:
            result = self._handle_missing_requirement(e)

        if result.result == scheduler.ST_TestResult.r_fatal:
            print self.formatter.output_one_result(test.dsc['header']['name'], result, self.options)
            raise tuierrors.SecToolActionError(self.name, _("Test %s failed in a fatal way - returned %d. Stopping execution of the rest of the tests" % (test.dsc['header']['name'], result.return_code)))

        return result

    def __email_test(self, test_name, result, diff):
        """
        Uses the MailOutput object to add info about one test ran to
        the email being composed.

        @param test_name: test name
        @type test_name: string
        @param result: test result
        @type result: ST_TestResult
        @param diff: result diff
        @type diff: ST_TestResultDiff
        """
        if self.scheduler_opts['MAIL']['SEND_BODY'].upper() == 'FULL':
            self.mail.add_result(test_name, result)
        elif self.scheduler_opts['MAIL']['SEND_BODY'].upper() == 'DIFF' and diff != None:
            self.mail.add_diff(test_name, result, diff)

        if self.scheduler_opts['MAIL']['SEND_ATTACHMENT'].upper() == 'FULL':
            self.mail.add_result(test_name, result, False)
        elif self.scheduler_opts['MAIL']['SEND_ATTACHMENT'].upper() == 'DIFF' and diff != None:
            self.mail.add_diff(test_name, result, diff, False)

    def __add_to_report(self, result):
        self.report.total += 1
        self.report[result.result] += 1

    def run_suite(self, suite, level, allow_default = False):
        """
        Iterates through the test suite given with the parameter suite,
        runs the test at level given with the parameter level and calls the
        appropriate formatter methods to produce output.

        @param suite: the suite to run tests from
        @type suite: ST_TestSuite
        @param level: level to run tests at
        @type level: integer
        @param allow_default: if set to True and test was to be run at a level
        which is not defined for the test, the method will retry on a default level
        @type allow_default: boolean
        """
        suite.updateEnviron(self.environ)

        # run the tests, print the results in a nice way
        for test in suite:
            result = scheduler.ST_TestResult()

            try:
                result = self.__run_one_test(test, level)
            except ValueError, e:
                if allow_default:
                    result = self.__run_one_test(test)
            except IOError, e:
                raise tuierrors.SecToolActionError(self.name, e)
            finally:
                if result.return_code != scheduler.ST_TestResult.NOT_YET_RAN:
                    # FIXME - problem - the xml is read and parsed upon completing every test, might be slow
                    diff = self.xml_output.write_result(test.dsc['header']['name'], result)

                    if self.options.report:
                        self.__add_to_report(result)

                    if self.options.mail:
                        self.__email_test(test.dsc['header']['name'], result, diff)

                    if self.options.diff and diff != None:
                        content = self.formatter.output_one_diff(test.dsc['header']['name'], result, diff, self.options)
                        if content != "": print content
                    else:
                        print self.formatter.output_one_result(test.dsc['header']['name'], result, self.options)

    def assert_suite_not_empty(self, suite):
        """
        Reports to stderr if a given test suite contained any tests 
        
        @param suite: the suite to check
        @type suite: ST_TestSuite
        @rtype: boolean
        """
        if suite.countTestCases() == 0:
            raise tuierrors.SecToolActionError(self.name, _("No matching tests to run"))

        return True

    def check_level_value(self, level):
        """
        Test if level number from args is between 1 and max_level variable

        @param level: the level to check
        @type level: integer
        @raises tuierrors.SecToolActionError
        """
        if level not in range(1, scheduler.ST_TestCase.max_level+1):
            raise tuierrors.SecToolActionError(self.name, _("Bad level specification - level must be between 1 and %d, got %d" % (scheduler.ST_TestCase.max_level, level)))

    def preReq(self):
        """
        Checks that user is root and creates the mail output object
        """
        ST_Action.preReq(self)
        self.checkRoot()

        self.xml_output = xmloutput.XmlOutput(os.path.join(self.environ['TDATA_DIR_BASE'], self.scheduler_opts['RESULTS']['RESULT_FILE']))
        if self.options.mail:
            self.mail = mailoutput.MailOutput(self.options, self.scheduler_opts)

    def tearDown(self):
        """
        Finishes composing the mail and sends it after the actual run is done
        """
        ST_Action.tearDown(self)

        if self.options.mail:
            self.mail.fini()
            retval = self.mail.send() 
            if retval != 0:
                raise tuierrors.SecToolActionError(self.name, _("Could not send mail: retcode = %d") % (retval))

        if self.options.report:
            self.formatter.output_summary(self.report.total,
                                          self.report[scheduler.ST_TestResult.r_passed],
                                          self.report[scheduler.ST_TestResult.r_warning],
                                          self.report[scheduler.ST_TestResult.r_error],
                                          self.report[scheduler.ST_TestResult.r_failure],
                                          self.report[scheduler.ST_TestResult.r_fatal])


class ST_CleanCache(ST_Action):
    """ Removing the cached test results"""
    name = _("remove old test results")

    def __init__(self, options, args, formatter=output.StdoutFormatter()):
        ST_Action.__init__(self, options.options, options.args, formatter)
        self.schedopt = options.scheduler_opts
        self.environ = dict([ (key.upper(), options.scheduler_opts['PATHS'][key]) for key in options.scheduler_opts['PATHS'] ])

    def doStuff(self):
        """ Actually removes the results file """
        ST_Action.doStuff(self)
        try:
            os.remove(os.path.join(self.environ['TDATA_DIR_BASE'], self.schedopt['RESULTS']['RESULT_FILE']))
        except OSError, e:
            if e.errno != errno.ENOENT:
                raise e

    def preReq(self):
        """ User must be root to purge the results """
        ST_Action.preReq(self)
        self.checkRoot()

class ST_RunLevel(ST_RunAction):
    """ Runs one whole security level """
    name = _("run test(s) at a certain level")

    def __init__(self, options, formatter=output.StdoutFormatter()):
        ST_RunAction.__init__(self, options, formatter)

    def _handle_missing_requirement(self, exception):
        result = scheduler.ST_TestResult()
        result.return_code = scheduler.ST_TestResult.NOT_YET_RAN

        return result

    def preReq(self):
        """ Prerequisite checks before the actual action """
        ST_RunAction.preReq(self)
        self.check_level_value(self.options.level)

    def doStuff(self):
        """ Actually runs the test level """
        ST_RunAction.doStuff(self)

        self.load_tests(self.environ['TESTS_DIRS'])

        # filter out by group and apply include/exclude
        self.filter_groups(self.suite, self.options.group)
        self.suite.filterInPlace(ST_Auto.relative_filter,
                                 level = self.options.level,
                                 add = self.options.include,
                                 remove = self.options.exclude )

        self.assert_suite_not_empty(self.suite) 
        self.run_suite(self.suite,
                       self.options.level,
                       (self.options.include != [] or\
                        self.options.exclude != []))

class ST_RunTests(ST_RunAction):
    """ Running a test or a collection of tests usually given by globbing """
    name = _("run test(s)")

    def __init__(self, options, formatter=output.StdoutFormatter()):
        ST_RunAction.__init__(self, options, formatter)

    def preReq(self):
        """ Prerequisite checks before the actual action """
        ST_RunAction.preReq(self)
        self.checkList()
        if self.options.level != scheduler.ST_TestCase.default_level:
            self.check_level_value(self.options.level)

    def doStuff(self):
        """ Actually runs the tests """
        ST_RunAction.doStuff(self)
        suite  = scheduler.ST_TestSuite()

        test_paths, positional = self.extractPathFromArgs(self.args)

        try:
            # load tests given by path
            suite.addTests( [ scheduler.ST_TestCase(self.loader.loadTest(test_path)) for test_path in test_paths ] )

            # treat the rest as given by name
            found = self.loader.loadFromPath(self.environ['TESTS_DIRS'])
        except exceptions.Exception, e:
            raise tuierrors.SecToolActionError(self.name, e)

        # filter by glob and add to the suite
        for glob in positional:
            suite.addTests( ( test for test in found if matches_glob(glob, test) ) )

        # filter out by group
        self.filter_groups(suite, self.options.group)

        # if we ended up with an empty suite, report that
        self.assert_suite_not_empty(suite) 
        self.run_suite(suite, self.options.level, allow_default=True)

class ST_InfoTests(ST_Action):
    """ Query info about test(s) """

    name = _("query info about test(s)")

    def __init__(self, options, formatter=output.StdoutFormatter()):
        ST_Action.__init__(self, options.options, options.args, formatter)
        self.path = options.scheduler_opts['PATHS']['TESTS_DIRS']

    def preReq(self):
        """ Prerequisite checks before the actual action """
        ST_Action.preReq(self)
        self.checkList()      # this command requires a test or a list of tests

    def doStuff(self):
        """ Actually prints info about the tests """
        ST_Action.doStuff(self)

        suite  = scheduler.ST_TestSuite()

        test_paths, positional = self.extractPathFromArgs(self.args)

        try:
            # load tests given by path
            suite.addTests( [ scheduler.ST_TestCase(self.loader.loadTest(test_path)) for test_path in test_paths ] )
            # treat the rest as given by name
            found = self.loader.loadFromPath(self.path)
        except exceptions.Exception, e:
            raise tuierrors.SecToolActionError(self.name, e)

        # filter by glob
        for glob in positional:
            suite.addTests( ( test for test in found if matches_glob(glob, test) ) )

        # filter out by group
        self.filter_groups(suite, self.options.group)

        # do the actual output
        self.formatter.output_test_info(suite)

class ST_ListTests(ST_Action):
    """ Lists know tests, supports globbing """
    name = _("list tests")

    def __init__(self, options, formatter=output.StdoutFormatter()):
        ST_Action.__init__(self, options.options, options.args, formatter)
        self.path = options.scheduler_opts['PATHS']['TESTS_DIRS']

    def preReq(self):
        """ Prerequisite checks before the actual action """
        ST_Action.preReq(self)

    def doStuff(self):
        """ Actually lists the tests """
        ST_Action.doStuff(self)

        self.load_tests(self.path)

        if len(self.args) == 0: # the simple case - we output them all
            filtered = self.suite
        else:                   # filter by args, each arg is a glob pattern
            filtered = scheduler.ST_TestSuite()
            for glob in self.args:
                filtered.addTests( ( test for test in self.suite if matches_glob(glob, test) ) )

        # filter out by group
        self.filter_groups(filtered, self.options.group)

        self.formatter.output_test_list(filtered)


class ST_Auto(ST_RunAction):
    """ Perform action defined in the confing file """
    name = _("action defined in the confing file")

    def __init__(self, options, formatter=output.StdoutFormatter()):
        ST_RunAction.__init__(self, options, formatter)
        self.environ = dict([ (key.upper(), options.scheduler_opts['PATHS'][key]) for key in options.scheduler_opts['PATHS'] ])

        try:
            options.scheduler_opts['ACTION']
        except KeyError:
            raise tuierrors.SecToolActionError(self.name, _("Please check if your config file contains a section 'ACTION'"))

        try:
            self.level = int(options.scheduler_opts['ACTION']['LEVEL'])
        except NoOptionError:
            self.level = ST_TestCase.default_level
        else:
            self.check_level_value(self.level)

        try:
            self.run  = options.scheduler_opts['ACTION']['RUN'].split()
        except NoOptionError:
            self.run = []

        self.relative = False

    def preReq(self):
        """ Prerequisite checks before the actual action """
        ST_RunAction.preReq(self)
        self.checkRoot()

    @staticmethod
    def name_filter(test, **kwargs):
        """ Filter method for self.suite - see filterInPlace documentation """
        return test.dsc['header']['name'] in kwargs['list']

    @staticmethod
    def relative_filter(test, **kwargs):
        level = kwargs['level']
        add = kwargs['add']
        remove = kwargs['remove']

        if level in test.dsc.valid_levels:
            if test.dsc['header']['name'] in remove:
                return False
            else:
                return True
        else:
            if test.dsc['header']['name'] in add:
                return True
            else:
                return False

        return False

    def parse_test_list(self):
        add = list()
        remove = list()
        blank = list()

        for tst in self.run:
            if tst.startswith('+'):
                add.append(tst[1:])
            elif tst.startswith('-'):
                remove.append(tst[1:])
            else:
                blank.append(tst)

        if add != [] or remove != []:
            self.relative = True

        if self.relative and blank != []:
            raise SecToolActionError(self.name, _("Cannot use both include/exclude and test name in test list"))

        return add, remove

    def doStuff(self):
        """ Actually runs the defined action """
        ST_RunAction.doStuff(self)

        add, remove = self.parse_test_list()

        self.load_tests(self.environ['TESTS_DIRS'])
        self.filter_groups(self.suite, self.options.group)

        if self.relative:
            self.suite.filterInPlace(self.relative_filter,
                                     level = self.level,
                                     add = add,
                                     remove = remove)
        else:
            if self.run != []:
                self.suite.filterInPlace(self.name_filter, list=self.run)

        self.assert_suite_not_empty(self.suite) 
        self.run_suite(self.suite, self.options.level, self.relative)

