# Copyright (c) 2007 Red Hat
# Written by Jakub Hrozek
# -*- coding: UTF8 -*-

from UserDict import UserDict

from scheduler.i18n import _
from scheduler.scheduler import OutputItem
from scheduler.scheduler import (ST_TestCase, ST_TestResult)

class RunReport(UserDict):
    def __init__(self):
        UserDict.__init__(self)
        self.total  = 0
        self[ST_TestResult.r_passed]     = 0
        self[ST_TestResult.r_warning]    = 0
        self[ST_TestResult.r_error]      = 0
        self[ST_TestResult.r_failure]    = 0
        self[ST_TestResult.r_fatal]      = 0


class OutputFormatter(object):
    """ A base class, meant to be subclassed. Defines several output methods, 
    one for every secTool action. The methods raise NotImplementedException
    by default, so these should not be called from the child class"""

    def output_one_diff(self, test_name, result, diff, options):
        """ 
        Outputs one test result diff

        @param test_name: name of the test
        @type test_name: string
        @param result: result of the test
        @type result: ST_TestResult
        @param diff: diff against previous run
        @type diff: ST_TestResultDiff
        @param options: tui options
        """
        raise NotImplementedError

    def output_one_result(self, test_name, results, options):
        """
        Outputs one test result

        @param test_name: name of the test
        @type test_name: string
        @param result: result of the test
        @type result: ST_TestResult
        @param options: tui options
        """
        raise NotImplementedError

    def output_test_list(self, tests):
        """ 
        Outputs a list of all known tests

        @param tests: the tests to list
        @type tests: ST_TestResult list
        """
        raise NotImplementedError

    def output_test_info(self, tests):
        """ 
        Outputs info about a test 
        @param tests: the tests to list
        @type tests: ST_TestResult list
        """
        raise NotImplementedError

    def output_summary(self, total, passed, warn, error, fail, fatal):
        """
        Outputs nice summary about all the tests that were run
        """
        raise NotImplementedError

class StdoutFormatter(OutputFormatter):
    """ Output which typically aims to a tty and will be read by a human. This
    is also secTool default. """

    # padding for listing results
    w_label = 12
    w_id    = 6
    w_total = 80

    # padding for listing tests
    list_w_name   = 15
    list_w_levels = 10 
    list_w_desc = w_total - list_w_name - list_w_levels

    # padding for info results
    info_pad = 12
    report_pad = 15

    str_test_results = [ _("PASS"), _("WARNING"), _("ERROR"), _("FAIL"), _("FATAL")]
    str_labels = { OutputItem.t_error : _("Error"),
                    OutputItem.t_warning : _("Warning"),
                    OutputItem.t_hint : _("Hint"),
                    OutputItem.t_info : _("Info") }

    def __get_diff_header(self, test_name, result_before, result_after):
        """
        Returns a nicely formatted test header

        @param test_name: name of the test
        @type test_name: string
        @param result_before: textual representation of the previous result (PASS, FAIL, ...)
        @type result_before: string
        @param result_after: textual representation of the current result (PASS, FAIL, ...)
        @type result_after: string
        """
        header = _("Test Name: %s") % (test_name)
        if result_before != result_after:
            str_result = (_("Test Result: %s (Was: %s)") % (result_after, result_before))
            str_result = str_result.rjust(self.w_total - len(header))
        else:
            str_result = (_("Test Result: %s") % (result_before)).rjust(self.w_total - len(header))
        header = "%s%s" % (header, str_result)

        return header

    def output_one_diff(self, test_name, result, diff, options):
        """
        See superclass definition

        @rtype: string
        """
        payload = ""
        if diff.are_same == True:
            if options.debug:
                payload += "%s\n" % self.__get_result_header(test_name, self.str_test_results[result.result])
                payload += _("\tThe test did not change its result\n")
            return payload

        payload = "%s\n" % (self.__get_diff_header(test_name,
                                 self.str_test_results[diff.before.result],
                                 self.str_test_results[diff.after.result]))

        for item in diff.added:
            msg = ("\t + %s(%s)\t" % (self.str_labels[item.type], item.msg_id)).ljust(15)
            msg += item.msg.lstrip()
            payload += "%s\n" % msg

        for item in diff.removed:
            msg = ("\t - %s(%s)\t" % (self.str_labels[item.type], item.msg_id)).ljust(15)
            msg += item.msg.lstrip()
            payload += "%s\n" % msg

        payload += '\n'
        return payload

    def __get_result_header(self, test_name, test_result):
        """
        Returns a nicely formatted test header

        @param test_name: name of the test
        @type test_name: string
        @param test_result: textual representation of the result (PASS, FAIL, ...)
        @type test_result: string
        """
        header = _("Test Name: %s") % (test_name)
        str_result = (_("Test Result: %s") % (test_result)).rjust(self.w_total - len(header))
        header = _("%s%s") % (header, str_result)

        return header

    def output_one_result(self, test_name, result, options):
        """
        See superclass definition

        @rtype: string
        """

        payload = ""

        # print the header including result severity
        test_result = self.str_test_results[result.result]
        payload += "%s\n" % self.__get_result_header(test_name, test_result)

        # output stderr if the test ended up with a failure or if the user specified the --debug option
        if result.result == ST_TestResult.r_fatal or result.result == ST_TestResult.r_failure or options.debug:
            for failure in result.stderr:
                msg  = _("\tFailure").ljust(15)
                msg += failure
                payload += msg.rstrip()

        # regular test output
        for item in result.output_items:
            # don't print hints without the proper option
            if item.type == OutputItem.t_hint and options.hint == False:
                continue

            # debug messages are printed without the id fiels
            if item.type == OutputItem.t_info:
                if options.debug == True:
                    msg = ("\t%s" % (self.str_labels[item.type])).ljust(15)
                else:
                    continue
            else:
                msg = ("\t%s(%s)" % (self.str_labels[item.type], item.msg_id)).ljust(15)

            # add the actual payload and print it
            msg += item.msg.strip()
            payload += "%s\n" % msg

        payload += '\n'
        return payload


    def __print_list_header(self):
        """ Prints a textual representation of header of sectool --list """
        print "%s\n%s %s %s\n%s" % ("-"*80,
                                   _("Name").ljust(self.list_w_name), 
                                   _("Description").ljust(self.list_w_desc), 
                                   _("Levels").ljust(self.list_w_levels),
                                   "-"*80)

    def __print_list_one_test(self, test_dsc):
        """ Prints one line of info about a test. """
        # if the description is too wide, append dots
        desc = test_dsc['header']['description'].ljust(self.list_w_desc)
        if '\n' in desc:
            desc = desc.replace('\n', ' ')

        if len(desc) > self.list_w_desc:
            desc = "%s..." % desc[:self.list_w_desc-3]

        # align corresponding levels
        str_levels = ""
        for i in xrange(1, ST_TestCase.max_level+1):
            if i in test_dsc.valid_levels:
                str_levels += "%d " % (i)
            else:
                str_levels += "  "

        # print as one string
        print "%s %s %s" % (test_dsc['header']['name'].ljust(self.list_w_name)[:self.list_w_name],
                            desc,
                            str_levels)

    def output_test_list(self, tests):
        """ See superclass for docs """
        self.__print_list_header()
        for tst in tests:
            self.__print_list_one_test(tst.dsc)

    def __print_info_one_test(self, test_dsc):
        """ Prints info about a test. """

        # FIXME - wrap description after 40 chars. Take a look at how yum does it.
        info_template = "%s : %s\n%s : %s\n%s : %s\n%s : %s\n%s : %s\n"
        print info_template % (_("Name").ljust(self.info_pad), test_dsc['header']['name'], 
                                _("Script name").ljust(self.info_pad), test_dsc['header']['file_name'],
                                _("Groups").ljust(self.info_pad), test_dsc['header']['groups'],
                                _("Levels").ljust(self.info_pad), test_dsc['header']['levels'],
                                _("Description").ljust(self.info_pad), test_dsc['header']['description'])

    def output_test_info(self, tests):
        """ See superclass for docs """
        for tst in tests:
            self.__print_info_one_test(tst.dsc) 

    def output_summary(self, total, passed, warn, error, fail, fatal):
        """ See superclass for docs """
        report_template = "%s : %s\n%s : %s\n%s : %s\n%s : %s\n%s : %s\n%s : %s\n"

        print "\n"
        print "----- %s -----" % (_("Run report"))
        print report_template % (_("Total tests run").ljust(self.report_pad), total,
                                 _("Passed").ljust(self.report_pad), passed,
                                 _("Warnings").ljust(self.report_pad), warn,
                                 _("Errors").ljust(self.report_pad), error,
                                 _("Failures").ljust(self.report_pad), fail,
                                 _("Fatal failures").ljust(self.report_pad), fatal)


class XmlFormatter(OutputFormatter):
    """ TBD """
    pass

class HtmlFormatter(OutputFormatter):
    """ TBD """
    pass
