# bash.defs
# secTool support functions for bash

shopt -s expand_aliases
alias const='declare -xr'
alias int='declare -xi'
alias var='declare -x'

# Global constants
const E_OK=0
const E_FAIL=1
const E_FATAL=2

const PASSWD_FILE='/etc/passwd'
const GROUP_FILE='/etc/group'
const SHADOW_FILE='/etc/shadow'

# gettext settings
const TEXTDOMAIN='sectool'
. gettext.sh

# Global functions

function DEBUGMSG () {
	echo $* 1>&2
}

#
# resolves dashes on the beginning of string
#
myeval_gettext () {
  gettext -- "$1" | (export PATH `envsubst --variables -- "$1"`; envsubst -- "$1")
}

function report {
    local type="$1"

    case "${type}" in
	WARNING) 
    		local id=$2
    		local text="$3"

    		local E_TEXT="`myeval_gettext \"${text}\"`"
		printf "W%02d : %s\n" "${id}" "${E_TEXT}"
	    ;;
	ERROR)   
    		local id=$2
    		local text="$3"

    		local E_TEXT="`myeval_gettext \"${text}\"`"
		printf "E%02d : %s\n" "${id}" "${E_TEXT}"
	    ;;
	HINT)    
    		local id=$2
    		local text="$3"

    		local E_TEXT="`myeval_gettext \"${text}\"`"
		printf "H%02d : %s\n" "${id}" "${E_TEXT}"
	    ;;
        INFO|DEBUG)
		local text="$2"
        	local E_TEXT="`myeval_gettext \"${text}\"`"
    		printf "I : %s\n" "${E_TEXT}"
            ;;
	*) DEBUGMSG "Bad parameter to report: ${type}"
	    ;;
    esac
}

# ?
function test_exit {
    if (( $1 == 0 )); then
	printf "Test: ${NAME}, Succeeded\n"
	return ${E_OK}
    else
	printf "Test: ${NAME}, Error: %d, - %s\n" "$1" "$2" 1>&2
	exit $1
    fi
}

function isset () {
    case "${1}" in
        [Yy][Ee][Ss]|[Tt][Rr][Uu][Ee]|1)
            return ${E_OK}
            ;;
        [Nn][Oo]|[Ff][Aa][Ll][Ss][Ee]|0)
            return ${E_FAIL}
            ;;
    esac
    return ${E_FAIL}
}

# <file> [1|0]
#         1 = hash only
#         0 = "SHA256(file) = hash"

function sha256file () {
    if isset ${2}; then
	openssl sha -sha256 "${1}" 2> /dev/null | sed -n 's/.*\([A-Fa-f0-9]\{64\}\).*/\1/p'
    else
	openssl sha -sha256 "${1}" 2> /dev/null
    fi
    return ${E_OK}
}

# <file> <perm> <user:group> [0|1 <desc>]
# 1 - verbose
# 0 - quiet

function check_file_perm () {
    if [[ -a "${1}" ]]; then
        if (( $(stat -c '%a' "${1}") > $2 )); then
            if (( ${4} == 1 )); then
                report 'WARNING' ${5} "Wrong permissions on $(stat -c '%F' "${1}"): \"${1}\" (${6:-unknown}, required permissions are ${2})"
            fi
            return ${E_FAIL}
        fi

        if ! (stat -c '%U:%G' "${1}" | grep -q "${3}"); then
            if (( ${4} == 1 )); then
                report 'WARNING' ${5} "Wrong owner/group on $(stat -c '%F' "${1}"): \"${1}\" (${6:-unknown}, required owner/group is ${3})"
            fi
            return ${E_FAIL}
        fi

        return ${E_OK}
    else
        if (( ${4} == 1 )); then
            report 'ERROR' ${5} "Missing file or directory: \"${1}\" (${6:-unknown})"
        fi
        return ${E_FAIL}
    fi
}

# <file> <perm> <user:group> [0|1 <desc>]
# (strict version)

function check_file_perm_s () {
    if [[ -a "${1}" ]]; then
        if (( $(stat -c '%a' "${1}") == ${2} )); then
            if (( ${4} == 1 )); then
                report 'WARNING' ${5} "Wrong permissions on $(stat -c '%F' \"${1}\"): \"${1}\" (${6:-unknown}, required permissions are ${2})"
            fi
            return ${E_FAIL}
        fi

        if ! (stat -c '%U:%G' "${1}" | grep -q "${3}"); then
            if (( ${4} == 1 )); then
                report 'WARNING' ${5} "Wrong owner/group on $(stat -c '%F' \"${1}\"): \"${1}\" (${6:-uknown}, required owner/group is ${3})"
            fi
            return ${E_FAIL}
        fi

        return ${E_OK}
    else
        if (( ${4} == 1 )); then
            report 'ERROR' ${5} "Missing file: \"${1}\" (${5:-unknown})"
        fi
        return ${E_FAIL}
    fi
}

# <uid>
function uid2user () {
    sed -n "s/^\([^:]*\):[^:]*:${1}:.*/\1/p" ${PASSWD_FILE}
}

# <gid>
function gid2user () {
    sed -n "s/^\([^:]*\):[^:]*:${1}:.*/\1/p" ${GROUP_FILE}
}

function rel2abs () {
    local DIR=$(dirname  ${1})
    local FIL=$(basename ${1})

    cd ${DIR}
    echo -n "`pwd`/${FIL}"
    cd ${OLDPWD}
}

# is it an ordinary user?
# takes one param - username
function canLogIn {
    if [[ "`egrep ^$1: ${PASSWD_FILE} | awk -F: '{ print $7 }'`" != "/sbin/nologin" ]]
    then 
	password="`egrep ^$1: ${SHADOW_FILE} | awk -F: '{ print $2 }'`"
	# length of passwd - very short or very long means invalid password and disabled account
	if [ "${#password}" -lt '13' ] || [ "${#password}" -gt '34'  ]
	then
	    return 0
	else
	    return 1
	fi
    else
	return 0
    fi
}
