#!/usr/bin/env sh

# ----------------------------------------------------------- #
# Copyright (C) 2008 Red Hat, Inc.                            #
# Written by Michel Samia <msamia@redhat.com>                 #
# home_dirs.sh                                                #
# more info in home_dirs.dsc                                  #
# ----------------------------------------------------------- #

# tests permissions and ownership of /home/*
# author: Michel Samia <msamia@redhat.com>

# test of /home/* for having same named user in passwd
# test of all 'real' users in passwd- if their directories exist and if they have right permissions
# test ownership of them


. "${TESTS_DIRS}/bash_defs.sh"

#constants
const passwd=/etc/passwd
const useradd=/etc/default/useradd
const default_homedir=/home

# --- Error IDs ------- #
const ID_HOMEDIR_CANTEXEC=1
const ID_HOMEDIR_CANTREAD=2
const ID_HOMEDIR_CANTWRITE=3
const ID_HOMEDIR_DOESNTEXISTS=4
const ID_HOMEDIR_GROUPOALLOWED=5
const ID_HOMEDIR_GROUPRALLOWED=6
const ID_HOMEDIR_GROUPWALLOWED=7
const ID_HOMEDIR_NOTDETECTED=8
const ID_HOMEDIR_NOUSER=9
const ID_HOMEDIR_OTHEROALLOWED=10
const ID_HOMEDIR_OTHERRALLOWED=11
const ID_HOMEDIR_OTHERWALLOWED=12
const ID_HOMEDIR_WRONGOWNER=13
const E_SAME_HOMES=14

#if[[ `whoami` -ne 'root' ]]
if [[ $UID -ne '0' ]]
then
    gettext "You have to be logged as root to run this test!" 1>&2 ; echo
    exit ${E_FAIL}
fi

####  tries to find default home directory ####
if [[ `egrep 'HOME=.+' $useradd|wc -l` -eq "1" ]]
then
    homedir=`egrep 'HOME=.+' $useradd|awk -F= '{print $2}'`
    #DEBUGMSG "Detected $homedir as users' home directory"
else
    report 'WARNING' $ID_HOMEDIR_NOTDETECTED "Could not detect users' home directory in \$useradd, the default will be used"  # "Could not detect users' home directory in $d"
    homedir=$defaulthomedir # not found, uses standard /home
fi

#### tests each directory, if there is the same named user in /etc/passwd ####
for i in $(find "${homedir}" -maxdepth 1 -mindepth 1 -type d -exec basename \{\} \;); do
    if [[ `grep  "^$i:.*:.*:.*:.*:$homedir/$i:.*$" $passwd|wc -l` -eq "0" && "$i" != "lost+found" ]]
    then
	report 'INFO' "Testing directory: $homedir/$i"
	report 'WARNING' $ID_HOMEDIR_NOUSER "Directory \$homedir/\$i doesn't have matching user in $passwd"!
	report 'HINT' $ID_HOMEDIR_NOUSER "It might be a lost directory after removed user. Check content of this directory and delete it, if it is no more used."
    fi
done

#### find users that don't have their home directory ####
#### and test permissions of existing home directories ####
{
i=0 # passwd line number

while read line
do
    i=$[ i+1 ]
    if [[ `echo $line| egrep ':/sbin/nologin$'|wc -l` -eq "0" ]] # user is 'normal' :-)
    then
	line_home="`echo $line | awk -F: '{print $6}'`"
	line_username="`echo $line | awk -F: '{print $1}'`"
	line_userID="`echo $line | awk -F: '{print $3}'`"
	
	if [[ "$line_username" == "sync" || "$line_username" == "shutdown" || "$line_username" == "halt" ]]
	then
	  continue
	fi

	report 'INFO' "testing user \"$line_username\""
	if ! [[ -d $line_home ]]
	then    
	    report 'WARNING' $ID_HOMEDIR_DOESNTEXISTS  "\$passwd line \$i: Home directory \"\$line_home\" doesn't exist!"
	    report 'HINT'    $ID_HOMEDIR_DOESNTEXISTS  "Please create this directory."
	else
	    perm=`stat -c '%A' $line_home`
	    ownerID=`stat -c '%u' $line_home`
	    ownername=`stat -c '%U' \$line_home`
	    
	    # ownership
	    if [[ ownerID -ne line_userID ]]
	    then
		report 'ERROR' $ID_HOMEDIR_WRONGOWNER "Home directory of user \"\$line_username\" is owned by user \"\$ownername\"!"
	    fi
	    
	    hint="Please set permissions 700 to his home directory!"
	    
	    # user
	    if [[ ${perm:1:1} != 'r' ]]
	    then
		report 'ERROR' $ID_HOMEDIR_CANTREAD "User \"\$line_username\" can't read his home directory!"
		report 'HINT' $ID_HOMEDIR_CANTREAD "$hint"
	    fi
	    if [[ ${perm:2:1} != 'w' ]]
	    then
		report 'ERROR' $ID_HOMEDIR_CANTWRITE "User \"\$line_username\" can't write to his home directory"
		report 'HINT' $ID_HOMEDIR_CANTWRITE "$hint"
	    fi
	    if [[ ${perm:3:1} != 'x' ]]
	    then
		report 'ERROR' $ID_HOMEDIR_CANTEXEC "User \"\$line_username\" can't execute his home directory!"
		report 'HINT' $ID_HOMEDIR_CANTEXEC "$hint"
	    fi

	    #group
	    if [[ ${perm:4:1} != '-' && "$line_username" != "root" ]]
	    then
		report 'ERROR' $ID_HOMEDIR_GROUPRALLOWED "User \"\$line_username\" allows other users from his group to read his home directory \"\$line_home\"!"
		report 'HINT' $ID_HOMEDIR_GROUPRALLOWED "$hint"
	    fi
	    if [[ ${perm:5:1} != '-' && "$line_username" != "root" ]]
	    then
		report 'ERROR' $ID_HOMEDIR_GROUPWALLOWED "User \"\$line_username\" allows other users from his group to write to his home directory \"\$line_home\"!"
		report 'HINT' $ID_HOMEDIR_GROUPWALLOWED "$hint"
	    fi
	    if [[ ${perm:6:1} != '-' && "$line_username" != "root" ]]
	    then
		report 'ERROR' $ID_HOMEDIR_GROUPOALLOWED "User \"\$line_username\" allows other users from his group to open his home directory \"\$line_home\"!"
		report 'HINT' $ID_HOMEDIR_GROUPOALLOWED "$hint"
	    fi
	    
	    # others
	    if [[ ${perm:7:1} != '-' ]]
	    then
		report 'ERROR' $ID_HOMEDIR_OTHERRALLOWED "User \"\$line_username\" allows other users to read his home directory \"\$line_home\"!"
		report 'HINT' $ID_HOMEDIR_OTHERRALLOWED "$hint"
	    fi
	    if [[ ${perm:8:1} != '-' ]]
	    then
		report 'ERROR' $ID_HOMEDIR_OTHERWALLOWED "User \"\$line_username\" allows other users to write to his home directory \"\$line_home\"!"
		report 'HINT' $ID_HOMEDIR_OTHERWALLOWED "$hint"
	    fi
	    if [[ ${perm:9:1} != '-' ]]
	    then
		report 'ERROR' $ID_HOMEDIR_OTHEROALLOWED "User \"\$line_username\" allows other users to open his home directory \"\$line_home\"!"
		report 'HINT' $ID_HOMEDIR_OTHEROALLOWED "$hint"
	    fi
	fi
    fi
done
}< $passwd


##### find duplicities in home directories #####

# first prepare only lines with real users - no daemons

echo "">/tmp/filtered_passwd
egrep "^.+:" $passwd |
{
    while read line
    do
	username="`echo $line | awk -F: '{ print $1 }'`"
	canLogIn "$username"
        if  [ $? -ne 0 ]
        then
           echo "$line">>/tmp/filtered_passwd
    	fi                                                                      
    done
}

cat /tmp/filtered_passwd | awk -F: '{ if ($1 != "") print $6 }' | sort | uniq -d |
{
    while read dir
    do
        lines="`grep -n -e "^.*:.*:.*:.*:.*:$dir:.*$" $passwd | awk -F: '{ printf "%s (user %s), ", $1, $2 }'| tr '\n' ','`" 
        lines="${lines%', '}" # remove last comma
        report 'ERROR' $E_SAME_HOMES "More users use the same home directory '\$dir' (\$passwd lines $lines)"
        report 'HINT' $E_SAME_HOMES "Please change home directories on these lines to be different or delete duplicate records"
    done
}
                                            
rm -f /tmp/filtered_passwd
